package org.netkernel.layer0.boot;

import java.io.File;
import java.io.FileInputStream;
import java.net.URI;
import java.util.*;
import java.util.Map.Entry;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.container.impl.Kernel;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.util.PropertyConfiguration;
import org.netkernel.layer0.util.Utils;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;

public class ModuleManager implements Runnable, IConfigurationListener
{
	private final IKernel mKernel;
	private final IModuleFactory[] mFactories;
	private Map<URI, ModuleWithLevel> mDeployedModules;
	private List<ModuleWithLevel> mBrokenModules;
	private List<URI> mBrokenURIs;
	private Map<URI,ModuleSourceWithLevel> mRegisteredModules;
	private Map<ModuleSourceWithLevel, ModuleSourceWithLevel> mReloadModules;
	
	private long mPollPeriod;
	private Thread mExpiryPollThread;
	private boolean mPaused;
	private boolean mStopped;
	private Set mToRemove;
	private long mSyncTime;
	private PropertyConfiguration mPropertyConfiguration;
	private IModuleManagerSyncListener mSyncListener;
	private int mRunLevel;
	private int mMaxRunLevel;
	
	private static ModuleManager sSingleton;
	
	public static final int DEFAULT_RUN_LEVEL=2;
	
	public ModuleManager(IKernel aKernel, IModuleFactory[] aFactories)
	{	mKernel=aKernel;
		mFactories=aFactories;
		ILogger logger = mKernel.getLogger();
		for(IModuleFactory f: mFactories)
		{	logger.log(ILogger.INFO, this, "MSG_BOOT_MODULE_FACTORY", f.getClass().getName());
		}
		mDeployedModules=new HashMap();
		mBrokenModules=new ArrayList();
		mBrokenURIs=new ArrayList();
		mRegisteredModules=new HashMap();
		mReloadModules=new HashMap();
		mToRemove=new HashSet();
		mRunLevel=0;
		mMaxRunLevel=mRunLevel;
		sSingleton=this;
		mKernel.addConfigurationListener(this);
	}
	
	public static ModuleManager getSingleton()
	{	return sSingleton;
	}
	
	public long getLastSyncTime()
	{	return mSyncTime;
	}
	
	public List<IModule> getModules()
	{	List<IModule> result = new ArrayList<IModule>(mDeployedModules.size()+mBrokenModules.size());
		for(Iterator<ModuleWithLevel> it=mDeployedModules.values().iterator(); it.hasNext();)
		{	result.add(it.next().getModule());
		}
		for (ModuleWithLevel mwl : mBrokenModules)
		{	result.add(mwl.getModule());
		}
		return result;
	}
	
	public Map<URI, ModuleSourceWithLevel> getRegisteredModules()
	{	return mRegisteredModules;		
	}
	
	public void setPropertyConfiguration(PropertyConfiguration aConfig)
	{	mPropertyConfiguration=aConfig;
	}
	
	public void configurationChanged(IConfiguration aConfiguration)
	{	mPollPeriod=aConfiguration.getLong("netkernel.poll", 0);

		if (mExpiryPollThread==null && mPollPeriod>0)
		{	mExpiryPollThread = new Thread(this,"Module Expiry Poll");
			mExpiryPollThread.setPriority(Thread.MIN_PRIORITY);
			mPaused=false;
			mStopped=false;
			mExpiryPollThread.start();
		}
		if (mExpiryPollThread!=null && mPollPeriod==0)
		{	mPaused=true;
			mStopped=true;
			mExpiryPollThread=null;
		}

	}	
	
	public void addModule(URI aSource)
	{	addModule(aSource, DEFAULT_RUN_LEVEL);
	}
	
	public void addModule(URI aSource, int aLevel)
	{	mRegisteredModules.put(aSource, new ModuleSourceWithLevel(aSource, aLevel));
	}
	
	public void removeModule(IModule aModule)
	{	mToRemove.add(aModule);
	}
	
	public void removeModulePermanently(IModule aModule)
	{	mToRemove.add(aModule);
		mRegisteredModules.remove(aModule.getMeta().getSource());
	}
	
	public boolean sync(boolean aFinal)
	{	
		boolean problems=false;
		mPaused=true;
		
		List<IModule> newModules=null;
		if (mKernel.getState()==IKernel.S_OPEN)
		{
			mKernel.setState(IKernel.S_CLOSED);
			
			Set toRemove = new HashSet();
			toRemove.addAll(mToRemove);
			
			//find all new modules that are going to be added
			newModules = new ArrayList<IModule>();
			for (Iterator<URI> i=mRegisteredModules.keySet().iterator(); i.hasNext(); )
			{	URI source = i.next();
				ModuleSourceWithLevel mswl=mRegisteredModules.get(source);
				if (mswl.getLevel()<=mRunLevel)
				{	if(!mDeployedModules.containsKey(source))
					{
						try
						{	IModule m = BootUtils.buildModule(mswl.getSource(),mKernel,mFactories);
							newModules.add(m);
						}	catch (Throwable e)
						{	// ignore - if it won't load then it won't have any dependencies
						}
					}
				}
			}
			//then use this list to detect which existing modules will be invalidated because of them
			for(Iterator<ModuleWithLevel> it=mDeployedModules.values().iterator(); it.hasNext();)
			{	ModuleWithLevel mwl=it.next();
				IModule m = mwl.getModule();
				if (m.isExpiredOnRestartWithNewModules(newModules))
				{	//classloader will be invalidated as a side-effect 
				}
			}
			
			do
			{	
				// do preDecommission
				for (Iterator i=toRemove.iterator(); i.hasNext(); )
				{	IModule m = (IModule)i.next();
					mDeployedModules.remove(m.getSource());
					mBrokenModules.remove(m);
					
					try
					{	m.preDecommissionModule();	
					}	catch (Exception e)
					{	logException(m.getSource(),e);
						problems=true;
					}
				}
				
				toRemove.clear();
				
				// now detect any side-effects through classloaders
				for(Iterator<ModuleWithLevel> it=mDeployedModules.values().iterator(); it.hasNext();)
				{	ModuleWithLevel mwl=it.next();
					IModule m = mwl.getModule();
					
					if (m.getSpaces().length>0)
					{
						ISpace space=m.getSpaces()[0];
						ClassLoaderWithExports cl=space.getClassLoader();
						if (cl!=null)
						{	if (cl.isInvalidated())
							{	toRemove.add(m);
								removeModule(m);
								ModuleSourceWithLevel mswl=new ModuleSourceWithLevel(m.getSource(),mwl.getLevel());
								mReloadModules.put(mswl,mswl);
							}
						}
					}
				}
			}
			while (!toRemove.isEmpty());
			
			mKernel.setState(IKernel.S_PAUSED);
		}

		for (Iterator i=mToRemove.iterator(); i.hasNext(); )
		{	IModule m = (IModule)i.next();
			try
			{	IModuleMeta meta = m.getMeta();
				mKernel.getLogger().log(ILogger.INFO, this, "MSG_BOOT_UNLOAD_MODULE",meta.getName(),meta.getVersion());
				m.decommisionModule(mKernel);	
			}	catch (Exception e)
			{	logException(m.getSource(),e);
				problems=true;
			}
		}
		mToRemove.clear();
		BootUtils.cleanupOnHotRestart();

		mSyncTime=System.currentTimeMillis();

		if (aFinal)
		{	mKernel.setState(IKernel.S_FINAL);
		}
		else
		{	
			//Decide the set to load
			HashSet<ModuleSourceWithLevel> toLoad=new HashSet();
			toLoad.addAll(mReloadModules.keySet());
			for (Iterator<URI> i=mRegisteredModules.keySet().iterator(); i.hasNext(); )
			{	URI source = i.next();
				ModuleSourceWithLevel mswl=mRegisteredModules.get(source);
				if(!toLoad.contains(mswl) && mswl.getLevel()<=mRunLevel)
				{	if(!mDeployedModules.containsKey(source))
					{	toLoad.add(mswl);
					}
				}
			}
			
			//Do the module load
			List<ModuleWithLevel> added=new ArrayList(mRegisteredModules.entrySet().size());
			List brokenURIs = new ArrayList();
			for(Iterator<ModuleSourceWithLevel>i=toLoad.iterator(); i.hasNext();)
			{	ModuleSourceWithLevel mswl=i.next();
				try
				{	IModule m=null;
					if (newModules!=null)
					{	//try to find in these modules we had to construct to check dependencies
						for (IModule nw: newModules)
						{	if (nw.getSource()==mswl.getSource())
							{	m=nw;
								break;
							}
						}
					}
					if (m==null)
					{	m	= BootUtils.buildModule(mswl.getSource(),mKernel,mFactories);
					}
					added.add(new ModuleWithLevel(m,mswl.getLevel()));
					mKernel.getLogger().log(ILogger.INFO, this, "MSG_BOOT_LOAD_MODULE",m.getMeta().getName(),m.getMeta().getVersion());
				}	catch (Throwable e)
				{	logException(mswl.getSource(),e);
					problems=true;
					mRegisteredModules.remove(mswl.getSource());
					brokenURIs.add(mswl.getSource());
				}
			}
			mReloadModules.clear();
			
			mBrokenURIs=brokenURIs;
			
			//Do the link
			for (Iterator<ModuleWithLevel> i=added.iterator(); i.hasNext(); )
			{	ModuleWithLevel mwl=i.next();
				IModule m = mwl.getModule();
				try
				{	m.linkModule(mKernel);
				}	catch (Exception e)
				{	i.remove();
					mBrokenModules.add(mwl);
					mRegisteredModules.remove(mwl.getModule().getSource());
					logException(m.getSource(),e);
					problems=true;
					try
					{	
						m.decommisionModule(mKernel);
					}	catch (Exception e2)
					{	logException(m.getSource(),e2);
					}
				}
			}
			
			mKernel.getLogger().log(ILogger.INFO, this, "MSG_BOOT_COMMISSION");
			
			//Do commission
			for (Iterator<ModuleWithLevel> i=added.iterator(); i.hasNext(); )
			{	ModuleWithLevel mwl=i.next();
				IModule m = mwl.getModule();
				try
				{	m.commissionModule(mKernel);
					mDeployedModules.put(mwl.getModule().getSource(),mwl);
				}	catch (Throwable e)
				{	i.remove();
					mBrokenModules.add(mwl);
					mRegisteredModules.remove(mwl.getModule().getSource());
					logException(m.getSource(),e);
					problems=true;
					try
					{	m.decommisionModule(mKernel);
					}	catch (Exception e2)
					{	logException(m.getSource(),e2);
					}
				}
			}

			configurationChanged(mKernel.getConfiguration());
			
			mKernel.setState(IKernel.S_READY);
			
			List<IModule> addedModules=new ArrayList<IModule>(added.size());
			for (ModuleWithLevel mwl : added )
			{	addedModules.add(mwl.getModule());
			}
			BootOrderOptimizer boo=BootOrderOptimizer.sInstance;
			boo.init(mKernel);
			boo.sort(addedModules);
			
			boo.reset(mKernel);
			
				
			for (IModule m : addedModules )
			{	try
				{	m.postCommissionModule();
				}	catch (Throwable e)
				{	logException(m.getSource(),e);
					problems=true;
				}
			}
			
			
			/*
			try
			{	long delay=mKernel.getConfiguration().getLong("stdmod.initialisation.hysteresis", 500L);
				Thread.sleep(delay);
			}
			catch (InterruptedException e)
			{	
			}
			*/
			
			mKernel.setState(IKernel.S_OPEN);
			mPaused=false;
		}
		
		return problems;
	}
	
	public void stop()
	{	mPaused=true;
		mStopped=true;
		
		for(Iterator<ModuleWithLevel> i=mDeployedModules.values().iterator(); i.hasNext();)
		{	ModuleWithLevel mwl=i.next();
			removeModule(mwl.getModule());
		}
		
		sync(true);
		sSingleton=null;
	}

	
	private void logException(URI aSource, Throwable aException)
	{	String msg = Utils.throwableToString(aException);
		String source;
		try
		{	source = aSource.toString()+"\n";
		}
		catch (NullPointerException e)
		{	source = "";
		}
		mKernel.getLogger().logRaw(ILogger.WARNING,this,source+msg);
	}

	private boolean mShouldMonitor=true;
	
	public void preventMonitoring()
	{	mShouldMonitor=false;
	}
	
	public void run()
	{	while(!mStopped)
		{	try
			{	if (mPollPeriod>0)
				{	Thread.sleep(mPollPeriod);
				}
				else
				{	break;
				}
				//Check init modules
				if(mShouldMonitor && !BootUtils.bootedFromJar(mKernel.getConfiguration()))
				{
					URI initmods=BootUtils.getInitModules(mKernel.getConfiguration());
					if(initmods.getScheme().equals("file"))
					{	URI base=URI.create(BootUtils.getInstallPath(mKernel.getConfiguration()));
						File f=new File(initmods);
						if(f.lastModified() > mSyncTime)
						{	IHDSNode mods=HDSFactory.parseXML(new FileInputStream(f));
							IHDSNodeList nl=mods.getNodes("//module");
							HashMap found=new HashMap(20);
							for(int i=0; i<nl.size(); i++)
							{	IHDSNode n=nl.get(i);
								URI u=base.resolve((String)n.getValue());
								found.put(u, u);
								ModuleWithLevel mod=mDeployedModules.get(u);
								String levelS=(String)n.getFirstValue("@runlevel");
								int level;
								if (levelS==null)
								{	level=DEFAULT_RUN_LEVEL;
								}
								else
								{	try
									{	level=Integer.parseInt(levelS);
									}
									catch(Exception e)
									{	mKernel.getLogger().logRaw(ILogger.WARNING, this, "Module runlevel not valid integer for "+u.toString());
										level=DEFAULT_RUN_LEVEL;
									}
								}
								if(mod==null)
								{	this.addModule(u,level);
								}
								else
								{	//look for changes in runlevel
									if (mod.getLevel()!=level)
									{	this.removeModule(mod.getModule());
										this.addModule(u, level);
									}
								}
							}
							//Now look for modules that are no longer here
							Set<URI>keys=mDeployedModules.keySet();
							for(Iterator<URI> i=keys.iterator(); i.hasNext(); )
							{	URI key=i.next();
								Object o=found.get(key);
								ModuleWithLevel mwl=mDeployedModules.get(key);
								if(o==null && mwl.getLevel()>1)
								{	this.removeModulePermanently(mDeployedModules.get(key).getModule());
								}
							}
							mKernel.getLogger().log(ILogger.INFO,this,"MSG_BOOT_CHANGES");
							sync(false);
						}
					}
				}
				
				//Check other modules
				if (!mPaused)
				{
					Set<ModuleWithLevel> changedSet=null;
					//look for directly changed modules
					for(Iterator<ModuleWithLevel> i=mDeployedModules.values().iterator(); i.hasNext();)
					{	ModuleWithLevel mwl=i.next();
						IModule m = mwl.getModule();
						if (m.isExpired(mSyncTime))
						{	if (changedSet==null)
							{	changedSet = new HashSet();
							}
						changedSet.add(mwl);
						}
					}
					for (Iterator<ModuleWithLevel> i=mBrokenModules.iterator(); i.hasNext(); )
					{	ModuleWithLevel mwl=i.next();
						IModule m =mwl.getModule();
						if (m.isExpired(mSyncTime))
						{	if (changedSet==null)
							{	changedSet = new HashSet();
							}
						changedSet.add(mwl);
						}
					}
					
					//look for changes to broken URIs
					for (URI uri : mBrokenURIs)
					{	if (uri.getScheme().equals("file"))
						{	File moduleDir = new File(uri);
							File[] files=moduleDir.listFiles();
							if (files!=null)
							{	for (File file : files)
								{	if (file.lastModified()>mSyncTime)
									{	if (changedSet==null)
										{	changedSet = new HashSet();
										}
										addModule(uri, mRunLevel);
										break;
									}
								}
							}
						}
					}
					
					if (changedSet!=null)
					{
						//apply changes
						mKernel.getLogger().log(ILogger.INFO,this,"MSG_BOOT_CHANGES");
						for (Iterator<ModuleWithLevel> i=changedSet.iterator(); i.hasNext(); )
						{	ModuleWithLevel mwl=i.next();
							IModule m = mwl.getModule();
							removeModule(m);
							//addModule(source);
							ModuleSourceWithLevel mswl=new ModuleSourceWithLevel(m.getSource(), mwl.getLevel());
							mReloadModules.put(mswl,mswl);
						}
						
						sync(false);
					}
					
					if (mPropertyConfiguration!=null)
					{	if (mPropertyConfiguration.isExpired())
						{	mPropertyConfiguration = new PropertyConfiguration(mPropertyConfiguration);
							((Kernel)mKernel).setConfiguration(mPropertyConfiguration);
						}
					}
				}
			}
			catch (Throwable t)
			{	logException(null,t);
			}
		}
	}
	
	public void registerSyncListener(IModuleManagerSyncListener aListener)
	{	mSyncListener=aListener;
	}
	
	public void notifySyncListener()
	{	if(mSyncListener!=null)
		{	mSyncListener.syncComplete();
		}
	}
	
	public void setRunLevel(int aLevel) throws IllegalArgumentException
	{	int lastLevel=mRunLevel;
		if(aLevel<0)
		{	throw new IllegalArgumentException("Level must be zero or positive integer");
		}
		if(aLevel>mRunLevel)
		{	mRunLevel=aLevel;
            mKernel.getLogger().log(ILogger.INFO,this,"MSG_RUNLEVEL_CHANGING",new Object[]{mRunLevel});
			stepUpRunLevel();
		}
		if(aLevel>0 && aLevel<mRunLevel)
		{	mRunLevel=aLevel;
            mKernel.getLogger().log(ILogger.INFO,this,"MSG_RUNLEVEL_CHANGING",new Object[]{mRunLevel});
			stepDownRunLevel();
		}
		if(mRunLevel>mMaxRunLevel)
		{	mMaxRunLevel=mRunLevel;			
		}
		if (lastLevel!=aLevel)
		{	mKernel.getLogger().log(ILogger.INFO,this,"MSG_RUNLEVEL_CHANGE",new Object[]{mRunLevel});
		}
	}
	
	private void stepUpRunLevel()
	{	sync(false);	
	}
	
	private void stepDownRunLevel()
	{	for(Iterator<ModuleWithLevel> i=mDeployedModules.values().iterator(); i.hasNext();)
		{	ModuleWithLevel mwl=i.next();
			if(mwl.getLevel()>mRunLevel)
			{	removeModule(mwl.getModule());				
			}
		}
		sync(false);
	}
	
	public int getRunLevel()
	{	return mRunLevel;		
	}
	
	public int getMaxRunLevel()
	{	return mMaxRunLevel;		
	}
}