package org.netkernel.layer0.urii;

import java.util.*;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.INKFParametersControl;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestReadOnlyImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.nkf.impl.ParametersImpl;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.*;
import org.netkernel.request.impl.*;
import org.netkernel.urii.*;
import org.netkernel.urii.impl.ResolutionImpl;


/** A space containing a set of supplied responses to SOURCE requests on given identifiers
 * @author Tony Butterfield
 */
public class ValueSpace implements ISpace
{
	private Object[] mData;
	private int mIndex;
	private IKernel mKernel;
	
	public static final IResponseMeta REQUEST_INDICATOR_META = new MetaImpl(ExpiryFactory.getAlwaysExpired());
	
	private static final IEndpoint sEndpoint = new ValueSpaceEndpoint();
	private static final RequestResponseFieldsImpl FIELDS=new RequestResponseFieldsImpl(1);
	static
	{	FIELDS.put(IResponseMeta.KEY_NO_CACHE, true);
	}
	public ValueSpace(int aSize)
	{	mData= new Object[aSize*3];
	}
	
	public ValueSpace(ValueSpace aOther)
	{	mIndex = aOther.mIndex;
		mData = new Object[aOther.mData.length];
		System.arraycopy(aOther.mData, 0, mData, 0, mIndex);
	}
	
	/** Put in a new value (not threadsafe but will only be initialised by a single thread anyway)
	 * @param aIdentifier identifier to match the source request
	 * @param aRepresentation representation in response
	 * @param aMeta optional meta - defaults to never expired
	 */
	public void put(String aIdentifier, Object aRepresentation, IResponseMeta aMeta)
	{	
		int putIndex=mIndex;
		for (int i=0; i<mIndex; i+=3)
		{	if (mData[i].equals(aIdentifier))
			{	putIndex=i;
				break;
			}
		}
		if (putIndex==mIndex)
		{	//resize if needed
			if (mIndex==mData.length)
			{	int newSize=mIndex*2;
				Object[] newData = new Object[newSize];
				System.arraycopy(mData, 0, newData, 0, mIndex);
				mData=newData;
			}
			mIndex+=3;
		}
		
		IResponseMeta meta = aMeta;
		if (meta==null)
		{	meta= MetaImpl.NEVER_EXPIRED;
		}
		mData[putIndex++]=aIdentifier;
		mData[putIndex++]=aRepresentation;
		mData[putIndex]=meta;
	}
	
	public void putRequest(String aIdentifier, NKFRequestImpl aRequest)
	{
	}
	
	public ClassLoaderWithExports getClassLoader()
	{	return null;
	}

	public void onLinkSpace(IKernel aKernel) throws Exception
	{
	}
	
	public void onCommissionSpace(IKernel aKernel) throws Exception
	{	mKernel=aKernel;
	}

	public void postCommissionSpace() throws Exception
	{	
	}
	
	public void preDecommissionSpace() throws Exception
	{
	}
 
	public void onDecommissionSpace() throws Exception
	{	
	}
	
	public boolean isCommissioned()
	{	return true;
	}
	
	protected IKernel getKernel()
	{	return mKernel;
	}
	
	protected IEndpoint getEndpoint()
	{	return sEndpoint;
	}

	public void onAsyncRequest(IRequest aRequest)
	{	
		IResponse result=null;
		int verb=aRequest.getVerb();
		
		if (verb==IRequest.VERB_RESOLVE)
		{	String identifier=aRequest.getIdentifier().toString();
			IRequest toResolve=(IRequest)aRequest.getValue(IRequest.TYPE_RESOLVE_REQUEST);
			IResolution resolution=null;
			verb=toResolve.getVerb();
			if (verb==IRequest.VERB_SOURCE || verb==IRequest.VERB_EXISTS)
			{	for (int i=0; i<mIndex; i+=3)
				{	if (mData[i].equals(identifier))
					{	resolution = new ResolutionImpl(sEndpoint,toResolve.getRequestScope());
						break;
					}
				}
			}
			result=new ResponseImpl(aRequest,resolution,MetaImpl.EXPIRED);
		}
		aRequest.getRequestor().onResponse(result);
	}
	
	IResponse getResponseFor(IRequest aRequest)
	{	IResponse result=null;
		int verb=aRequest.getVerb();
		if (verb==IRequest.VERB_SOURCE)
		{	String identifier=aRequest.getIdentifier().toString();
			for (int i=0; i<mIndex; i+=3)
			{	if (mData[i].equals(identifier))
				{	Object representation = mData[i+1];
					IResponseMeta meta = (IResponseMeta)mData[i+2];
					result=customHandler(aRequest,representation,meta);					
					if (result==null)
					{	result = new ResponseImpl(aRequest,representation,meta);
					}
					break;
				}
			}
		}
		else
		{	result = new ResponseImpl(aRequest,Boolean.TRUE, new MetaImpl(ExpiryFactory.getNeverExpired(),FIELDS));
		}
		return result;
	}
	
	protected IResponse customHandler(IRequest aRequest, Object aRepresentation, IResponseMeta aMeta )
	{	IResponse result=null;
		if (aMeta==REQUEST_INDICATOR_META)
		{	NKFRequestImpl req=(NKFRequestImpl)aRepresentation;
			NKFRequestReadOnlyImpl reqIn = new NKFRequestReadOnlyImpl(aRequest);
			INKFParametersControl params=new ParametersImpl();
			NKFContextImpl ctx = new NKFContextImpl(mKernel,this,reqIn,sEndpoint,params);
			IRequestScopeLevel scope=aRequest.getParent().getRequestScope();
			ctx.setRequestScope(scope); //use requestscope of parent request
			req.setContext(ctx);
			try
			{	INKFResponseReadOnly resp=ctx.issueRequestForResponse(req);
				NKFResponseImpl respOut=new NKFResponseImpl(resp, ctx);
				result = respOut.getKernelResponse();
			}
			catch (NKFException e)
			{	result = new ResponseImpl(aRequest, e, new MetaImpl(ExpiryFactory.getDependentExpiry()));
			}
		}
		return result;
	}
	
	public String toString()
	{	StringBuffer sb=new StringBuffer(128);
		sb.append("ValueSpace(");
		for (int i=0; i<mIndex; i+=3)
		{	if (i!=0)
			{	sb.append(", ");
			}
			sb.append(mData[i]);
		}
		sb.append(")");
		return sb.toString();
	}

	public boolean equals(Object aOther)
	{	boolean result=aOther==this;
		if (!result && aOther instanceof ValueSpace)
		{	ValueSpace other=(ValueSpace)aOther;
			result=mIndex==other.mIndex;
			if (result)
			{	Object[] od=other.mData;
				for (int i=0; i<mIndex; i+=3)
				{	if (mData[i].equals(od[i]))
					{	Object d1=mData[i+1];
						Object d2=od[i+1];
						if (d1==null || d2==null)
						{	if (d1!=d2)
							{	result=false;
								break;
							}
						}
						else if (!d1.equals(d2))
						{	result=false;
							break;
						}
					}
					else
					{	result=false;
						break;
					}
				}
			}
		}		
		return result;
	}
	
	/* Must be implemented so that equal spaces have equal hashCode for cache to work effectively
	 */
	public int hashCode()
	{	int hash=0;
		for (int i=0; i<mIndex; i+=3)
		{	hash^=mData[i].hashCode();
		}
		return hash;
	}

	/* Only resolvesEqual when all keys are the same
	 * @see org.netkernel.urii.ISpace#resolveEquals(java.lang.Object)
	 */
	public boolean resolveEquals(Object aOther)
	{	boolean equal=aOther instanceof ValueSpace;
		if (equal)
		{	ValueSpace pbvc=(ValueSpace)aOther;
			equal = (pbvc.mIndex==mIndex);
			if (equal)
			{	for (int i=0; i<mIndex; i+=3)
				{	if (!mData[i].equals(pbvc.mData[i]))
					{	equal=false;
						break;
					}
				}
			}
		}
		return equal;
	}
	
	/* following methods added to support NKP */
	public int getSize()
	{	return mIndex/3;
	}
	public String getIdentifier(int aIndex)
	{	return (String)mData[aIndex*3];
	}
	public Object getRepresentation(int aIndex)
	{	return mData[aIndex*3+1];
	}
	public IResponseMeta getResponseMeta(int aIndex)
	{	return (IResponseMeta)mData[aIndex*3+2];
	}
	
	public String createPBVIdentifier(String aName)
	{	String head="pbv:"+aName;
		int c=0;
		for (int i=0; i<mIndex; i+=3)
		{	String identifier=(String)mData[i];
			if (identifier.startsWith(head)) c++;
		}
		String result;
		if (c==0)
		{	result=head;
		}
		else
		{	result=head+c;
		}
		return result;
	}		
	
}
